//
// $Id: xy.h 101 2008-07-10 09:20:12Z mrchip $
//
// Copyright (C) 2005 STADT-LAND-FLUSS INGENIEURDIENSTE GmbH
//
// Authors:
// Ulrich Kiel <u.kiel@S-L-F.de>
//
// This program is free software under the GPL (>=v2)
// Read the file COPYING coming with WSPLGEN for details.
//

//---------------------------------------------------------------------------
#ifndef XYH
#define XYH

//----------------------------------------------------------------------------
#include <string>
#include <vector>
#include <list>
#include <set>
#include <map>

#include <math.h>

#include "quadtree.h"

//----------------------------------------------------------------------------
typedef enum { NO_ELEMENT, TRI, QUAD, NEWTRI } TElementTyp;
typedef enum { NOT_INSIDE, INSIDE, ON_LINE, UNDEFINED } TInsideTyp;

const int NoWsp = -999999;
const int NoZ = -1000000;
const int GrabenZ = -999999;
const int DammZ = 999999;
const int NotValid = -999999;

//----------------------------------------------------------------------------
class TXYZ;
class TNode;
class TPoint;
class TPointCompMeter;
// typedef set TPointList;
class TElement;
class TEdge;
class TEdgeIndex;
class TXYZList;
class TErgebnisPolygon;
class TNodeList;
class TElementList;
class TEdgeCompNr;
class TEdgeCompdistance;
class TEdgeList;
class TErgebnisPolygonList;
class TEdgeIndex;
class TProfil;
class TProfilCompStation;
class TProfilList;
// typedef TProfilMap
class TBereichCompStation;
class TBereichsList;
class TGewaesserAchse;
class TGewaesserAchseComp;
class TGewaesserAchseList;

//----------------------------------------------------------------------------
static const unsigned int 	MaxXYKombiIndexAnzahlKonst = 10;

//----------------------------------------------------------------------------
// TXY wird fr Sttzstellen von Polygonen, Knoten und Profilspuren bentigt.
// Hier werden reine X,Y-Koordinaten gespeichert.
//----------------------------------------------------------------------------
class TXYZ
{
	public:
		long  		X;
		long  		Y;
		int			Z;

		TXYZ(long X, long Y, int Z=0);
		TXYZ(TXYZ *Xyz);
};

//----------------------------------------------------------------------------
// TNode wird fr die Knoten in vermaschten Netzen genutzt
// Die Speicherung erfolgt im Allgemeinen in einer TNodeList
//----------------------------------------------------------------------------
class TNode : public TXYZ
{
	public:
		int			Nr;
		int 		Wsp;
		TEdgeIndex 	*EdgeIndex;

					TNode(int Nr, long X, long Y, int Z, int Wsp = NoWsp);
					TNode(TNode *Node);
					~TNode(void);

		void 		ClearEdgeIndex(void);
		void 		EraseEdgeIndex(TEdge *Edge);
};


//----------------------------------------------------------------------------
// TPoint wird fr die Sttzstellen von Profilspuren und anderen Spuren genutzt,
// die interpoliert oder verdichtet werden sollen.
//----------------------------------------------------------------------------
class TPoint : public TXYZ
{
	public:
		int  	Meter;

		TPoint(long X, long Y, int Z, int Meter);
		TPoint(TPoint *Point);
};

//----------------------------------------------------------------------------
// Hilfsklasse fr das automatische Sortieren von 'Points' nach Metern
// Wird von der TPointList bentigt
//----------------------------------------------------------------------------
class TPointCompMeter
{
	public:
		bool operator()(TPoint* const &Point1, TPoint* const &Point2) const
		{
			return (Point1->Meter < Point2->Meter);
		}
};

//----------------------------------------------------------------------------
// Hier werden die TPoints drin verwaltet
// durch die Verwendung von 'set' ist immer gewhrleistet, dass keine Punkte doppelt
// sind und die Punkte immer nach Metern sortiert sind
//----------------------------------------------------------------------------
typedef std::set<TPoint *, TPointCompMeter>			TPointList;

//----------------------------------------------------------------------------
// Diese Klasse speichert Drei- oder Vierecke (wobei die Vierecke noch nicht
// berall im Programm untersttzt werden
// Schn wre es, wenn man diese Klasse komplett wegoptimieren wrde
//
// !!! Achtung !!! Die Pointer auf die Knoten sind nur geliehen.
// Der Speicher der Knoten wird ber eine TNodeList verwaltet und darf nicht
// von TElement freigegeben werden.
//----------------------------------------------------------------------------
class TElement
{
	public:
		TElementTyp	Typ;
		TNode*		Node1;
		TNode*		Node2;
		TNode*		Node3;
		TNode*		Node4;

					TElement(TNode *Node1, TNode *Node2, TNode *Node3);
					TElement(TNode *Node1, TNode *Node2, TNode *Node3, TNode *Node4);
					TElement(TElement *Element);
};

//----------------------------------------------------------------------------
// TEdge dient dem Verwalten von Kanten
//
// !!! Achtung !!! Die Pointer auf die Knoten sind nur geliehen.
// Der Speicher der Knoten wird ber eine TNodeList verwaltet und darf nicht
// von TEdge freigegeben werden.
//----------------------------------------------------------------------------
class TEdge
{
	public:
		TNode 	*Node1;
		TNode 	*Node2;
		double  Distance;
		double	Arc;
		long	X;
		long	Y;
		bool	Ready;
		bool	IsBoundary;

				TEdge(TNode *Node1, TNode *Node2);
				TEdge(TEdge *Edge);
};

//----------------------------------------------------------------------------
// TXYList verwaltet die TXYZ
//----------------------------------------------------------------------------
class TXYZList : public std::vector<TXYZ *>
{
	private:
		TQuadTree<TXYZList, TXYZ>	*QuadTree;

	public:
		long		 	MinX;
		long		 	MaxX;
		long		 	MinY;
		long		 	MaxY;

						TXYZList(void);
						~TXYZList(void);
		void 			Clear(void);
		TXYZList* 		Copy(void);

		void 			Add(TXYZ *Xyz);
		void 			Add(long X, long Y, int Z = 0);
		void 			SortByXY(void);
		TXYZ* 			FindByXY(long X, long Y);
		TInsideTyp		IsInsideXYList(long X, long Y);
		TXYZList* 		Cut(TXYZList* ProfilPolygon, TProfil* CutProfil);
};

//----------------------------------------------------------------------------
// TnodeList dient zum Verwalten der Knoten (TNode)
// Nur TNodeList gibt den Speicher fr die Knoten wieder frei
//----------------------------------------------------------------------------
class TNodeList : public std::vector<TNode *>
{
	private:
		TQuadTree<TNodeList, TNode>*	QuadTree;
		TNodeList*						SearchNodeList;

	public:
		int		 			MaxNr;
		long		 		MinX;
		long		 		MaxX;
		long		 		MinY;
		long		 		MaxY;

		long		 		DiffX;
		long		 		DiffY;

							TNodeList(void);
							~TNodeList(void);
		void 				Clear(void);
		TNodeList* 			Copy(void);

		void				CreateQuadTree(long MinX, long MinY, long MaxX, long MaxY);
		void 				Add(TNode *Node);
		bool 				AddWithQuadTree(TNode *Node);
		bool 				AddUnique(TNode *Node);
		void				SortByXY(void);
		void				SortByNr(void);
		TNode*				FindByXY(long X, long Y);
		TNodeList*			FindAllByXY(double X, double Y, double R);
		TNode*				FindByNr(int Nr);
		bool				Interpolate(TProfilList* SperrenList, long X, long Y, double R, int *Z, int *Wsp);
		void				ClearEdgeIndex(void);
		int					GetMaxDepth(int DebugLevel);
		int 				GetMaxWsp(int DebugLevel);
		int 				GetMaxZ(int DebugLevel);
		void 				Scale(void);
		void 				ReScale(void);
};

//----------------------------------------------------------------------------
// TElementList verwaltet die Elemente (TElement)
//----------------------------------------------------------------------------
class TElementList : public std::vector<TElement *>
{
	private:
		int			HelpChangeElements(TNodeList *NodeList, TProfilList *LinienList, int DebugLevel);

	public:
					~TElementList(void);
		void		Clear(void);
		void		Add(TElement* Element);
		void		ChangeElements(TNodeList* NodeList, TProfilList *SperrenList, TProfilList *BruchList, int DebugLevel);
		void 		CleanChangedElements(int DebugLevel);
};


//----------------------------------------------------------------------------
// Ein 'set', das die Kanten nach Nummern sortiert speichert
//----------------------------------------------------------------------------
typedef std::vector<TEdge *>				TEdgeVector;

//----------------------------------------------------------------------------
// Die EdgeList besteht aus zwei internen 'sets', die fr einen schnellen
// Zugriff auf die Kanten dienen
//----------------------------------------------------------------------------
class TEdgeList
{
	private:
		// Diese beiden Knoten und die Kante dienen nur als Hllen fr das Suchen
		// damit das aufrufende Programm sich nicht erst immer wieder Speicher holen muss
		// und intern immer die gleichen Suchroutinen benutzt werden knnen
		TNode *SearchNode1;
		TNode *SearchNode2;
		TEdge *SearchEdge;

	public:
									TEdgeList(void);
									~TEdgeList(void);

		TEdgeVector					EdgeVector;

		unsigned int				Anz;
		double						AvgDistance;

		void 						Clear(void); // Wird verwendet, wenn auch die Kanten selbst gelscht werden sollen
		void 						Empty(void); // Wird verwendet, wenn es es nur geliehene Kanten waren
		void 						Add(TEdge *Edge);	// Wird nur fr Kanten benutzt, die nicht gefunden werden sollen (siehe Triangulate)
		TEdge* 						AddWithIndex(TNode *Node1, TNode *Node2);
		TEdge* 						FindWithIndex(TNode *Node1, TNode *Node2);

		void 						BuildDistanceSortedList(void);
		void 						BuildEdgeIndex(void);
};

//----------------------------------------------------------------------------
// Ein spezielles Polygon zum Speichern der Ergebnispolygone der Nass/Trockenberechnung
//----------------------------------------------------------------------------
class TErgebnisPolygon : public TXYZList
{
	public:
		double		Diff;

					TErgebnisPolygon(double Diff);

		long double	Area(void);
};

//----------------------------------------------------------------------------
// List von Ergebnispolygonen
//----------------------------------------------------------------------------
class TErgebnisPolygonList : public std::vector<TErgebnisPolygon *>
{
	public:
								~TErgebnisPolygonList(void);
		void 					Clear(void);
		TErgebnisPolygon*		First(void);
		TErgebnisPolygonList* 	Copy(void);
		void 					Append(TErgebnisPolygonList *AddErgebnisPolygonList);
		TInsideTyp				IsInsidePolygonList(long X, long Y);
};

//----------------------------------------------------------------------------
// Dieser Index schafft eine Schnelle Verbindung zwischen Knoten und
// Kanten (genau in diese Richtung und nicht von Kanten zu Knoten)
//----------------------------------------------------------------------------
class TEdgeIndex
{
	public:
		TEdge* 		Edge;
		TEdgeIndex* NextEdgeIndex;

		TEdgeIndex(TEdgeIndex* EdgeIndex, TEdge* Edge);
};

//----------------------------------------------------------------------------
// Hier werden Linienobjekte gespeichert
// Profile, aber auch Sperren und Brche
// und kurzfristig auch mal Wasserstnde
//----------------------------------------------------------------------------
class TProfil
{
	public:
		long		 	MinX;
		long		 	MaxX;
		long		 	MinY;
		long		 	MaxY;
		int				Station;

		TPointList*		PointList;

		long			Wsp;

						TProfil(int Station, int Wsp = 0);
						TProfil(TProfil *Profil);
						~TProfil(void);

		void  			AddPoint(int Meter);
		void  			AddPoint(long X, long Y, int Z = 0);
		bool			GetXY(int Meter, long *X, long *Y);
		void 			EqualizeProfil(TProfil *Profil);
		void			InterpolateProfil(TProfil *VorProfil, TProfil *NachProfil, double Prozent);
		bool			CalcSchnitt(TGewaesserAchseList *GewaesserAchseList, double* X, double *Y);
		bool			CalcSchnitt(TEdge *Edge);
		double  		Length(TPoint *Point1, TPoint *Point2);
		double  		Length(long X1, long Y1, long X2, long Y2);
		TPoint*			FindNearest(long X, long Y);
};

//----------------------------------------------------------------------------
// Hilfsklasse zum automatischen Sortieren nach Stationen (Flusskilometern)
//----------------------------------------------------------------------------
class TProfilCompStation
{
	public:
		bool operator()(TProfil* const &Profil1, TProfil* const &Profil2) const
		{
			return (Profil1->Station < Profil2->Station);
		}
};

//----------------------------------------------------------------------------
// TProfilList verwaltet die Profile (TProfil)
//----------------------------------------------------------------------------
class TProfilList : public std::set<TProfil *, TProfilCompStation>
{
	public:
					~TProfilList(void);

		void		Clear(void);
		TProfil* 	Find(int Station);
		void		InterpoliereProfile(double SollAbstand);
		void		FillProfile(double SollAbstand, int SollAnz, int DebugLevel);
		double		GetMaxWsp(int DebugLevel);
};

//----------------------------------------------------------------------------
// Hier wird eine Gewaesserachse gespeichert
// Eine Punktliste und ein paar Zusatzinformationen
//----------------------------------------------------------------------------
class TGewaesserAchse
{
	public:
//		int				ID;

		long		 	MinX;
		long		 	MaxX;
		long		 	MinY;
		long		 	MaxY;

		TPointList*		PointList;

						TGewaesserAchse(void);
						TGewaesserAchse(TGewaesserAchse *GewaesserAchse);
						~TGewaesserAchse(void);

		void  			AddPoint(int Meter);
		void  			AddPoint(long X, long Y, int Z = 0);
		bool			GetXY(int Meter, long *X, long *Y);
		double  		Length(TPoint *Point1, TPoint *Point2);
		double  		Length(long X1, long Y1, long X2, long Y2);
		TPoint*			FindNearest(long X, long Y);
};

//----------------------------------------------------------------------------
// Hier werden die Gewsserachsen drin verwaltet
//----------------------------------------------------------------------------
class TGewaesserAchseList : public std::vector<TGewaesserAchse *>
{
	public:
				~TGewaesserAchseList(void);

		void	Clear(void);
};

#endif

